#include "iodefine.h"
#include "bldc.h"

/* There are three users of the encoder:

   First, the linear position of the motor relative to it's starting
   point is encoder_lval() which can be zero'd by encoder_reset().
   This value is independent of the absolute angle of the rotor.

   Second, encoder_ang() returns a value of 0..1999 that reflects the
   rotor's absolute angle, relative to the phase angle.  I.e. it
   reflects the phase in the U-V-W phase.  This value may be an
   estimate until the zero pulse is seen.

   Third, encoder_rpm() returns the RPM of the motor, based on the
   last encoder pulse.  The range of RPMs that are supported is
   approximately 1 to 75,000 RPM.  In the 1000 RPM range, error is
   0.3%.  RPM is updated no less often than about 40 times a second
   (slowest reportable encoder rate)

   MTU0.TGRC = pclk clocks per encoder clock
   MTU1.TCNT = raw encoder clock
   MTU1.TGRA = encoder clock as of last zero pulse
*/


/* 48 bits at 4000 steps per rotation is 70 BILLION rotations, or
   about 22 years at 6,000 RPM.  If that's not enough, change this to
   "long long" :-)  */
static volatile int val_high_word = 0;

static volatile char zero_seen = 0;

static void __attribute__((interrupt))
overflow_handler ()
{
  val_high_word ++;
  (void)MTU1.TSR.BYTE;
  MTU1.TSR.BIT.TCFV = 0;
}

static void __attribute__((interrupt))
underflow_handler ()
{
  val_high_word --;
  (void)MTU1.TSR.BYTE;
  MTU1.TSR.BIT.TCFU = 0;
}

static void __attribute__((interrupt))
zero_handler ()
{
  zero_seen = 1;
  (void)MTU1.TSR.BYTE;
  MTU1.TSR.BIT.TGFA = 0;
  MTU1.TIER.BIT.TGIEA = 0;
}

#define SPEED_CLK (40000000/16)

int
encoder_rpm ()
{
  int tgrc = (uint16_t)MTU0.TGRC;
  if (tgrc == 0xffff)
    return 0;
  /* tgrc is ticks per enc half-phase */
  /* (tgrc*2000) is ticks per rot (MTU0 clocks off only one phase) */
  /* SPEED_CLK is ticks per sec */
  /* SPDC/(tgrc*2000) is thus rot/sec */
  return SPEED_CLK * 60 / (tgrc * 2000);
}

static void __attribute__((interrupt))
rpm_overflow_handler ()
{
  MTU0.TGRC = 0xffff;
  (void)MTU0.TSR.BYTE;
  MTU0.TSR.BIT.TCFV = 0;
}

void
encoder_init ()
{
  MTU1.TMDR.BIT.MD = 4; /* phase counting mode 1 */
  MTU1.TIOR.BIT.IOA = 8; /* Table 15.15: Input capture CNT->GRA on MTIOC1A */
  MTU.TSTRA.BIT.CST1 = 1; /* start MTU1 */

  MTU0.TMDR.BIT.MD = 0;	/* normal mode */
  MTU0.TMDR.BIT.BFA = 1; /* TGRA->TGRC buffered */
  /*  MTU0.TCR.BIT.TPSC = 1;*/ /* PCLK/4 */
  MTU0.TCR.BIT.TPSC = 2; /* PCLK/16 */
  MTU0.TCR.BIT.CCLR = 1; /* Table 17.6: clear on TGRA capture */
  MTU0.TIORH.BIT.IOA = 12; /* Table 15.13: TGRA input capture on MTU1 clock */
  MTU.TSTRA.BIT.CST0 = 1; /* start MTU0 */

  _vectors[VECT(MTU1,TCIV1)] = overflow_handler;
  _vectors[VECT(MTU1,TCIU1)] = underflow_handler;
  _vectors[VECT(MTU1,TGIA1)] = zero_handler;
  MTU1.TIER.BIT.TCIEU = 1;
  MTU1.TIER.BIT.TCIEV = 1;
  MTU1.TIER.BIT.TGIEA = 1;
  IEN(MTU1,TCIV1) = 1;
  IPR(MTU1,TCIV1) = 5;
  IEN(MTU1,TCIU1) = 1;
  IPR(MTU1,TCIU1) = 5;
  IEN(MTU1,TGIA1) = 1;
  IPR(MTU1,TGIA1) = 5;

  _vectors[VECT(MTU0,TCIV0)] = rpm_overflow_handler;
  MTU0.TIER.BIT.TCIEV = 1;
  IEN(MTU0,TCIV0) = 1;
  IPR(MTU0,TCIV0) = 5;

  /* Set an approximate "zero" based on the halls.  Close enough for
     inefficient commutation.  */
  MTU1.TCNT = 0;
  switch (HALLS & (U_HALL | V_HALL | W_HALL))
    {
    case U_HALL | V_HALL:	MTU1.TGRA =  -166; break;
    case V_HALL:		MTU1.TGRA =  -500; break;
    case V_HALL | W_HALL:	MTU1.TGRA =  -833; break;
    case W_HALL:		MTU1.TGRA = -1166; break;
    case U_HALL | W_HALL:	MTU1.TGRA = -1500; break;
    case U_HALL:		MTU1.TGRA = -1833; break;
    default:
      tty_puts ("invalid hall inputs\n");
    }
}

long long
encoder_lval ()
{
  unsigned short count;
  int r;

  do {
    r = val_high_word;
    count = MTU1.TCNT;
  } while (r != val_high_word);

  return ((long long)r << 16) | count;
}

int
encoder_val ()
{
  short count = MTU1.TCNT;
  short pos;
  short tgra = (short)(MTU1.TGRA);

  pos = count - tgra;
  pos %= 4000;
  if (pos < 0)
    pos += 4000;

  return pos;
}

int
encoder_ang ()
{
  short pos;
  int deg;

  pos = (short)MTU1.TCNT - (short)MTU1.TGRA;
  pos -= config_wiring.encoder_at_u;
  pos *= config_wiring.sincos_negate;
  pos %= 2000;
  if (pos < 0)
    pos += 2000;

  return pos;
}

int
encoder_synced ()
{
#if 0
  short count = MTU1.TCNT;
  int pos;
  short tgra = (short)(MTU1.TGRA);

  pos = count - tgra;
  if (pos < 0)
    pos += 4000;

  if ((unsigned)pos > 4010)
    return 0;
  return 1;
#else
  return zero_seen;
#endif
}

encoder_reset ()
{
  MTU1.TCNT -= MTU1.TGRA;
  val_high_word = 0;
}
