#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "bldc.h"

static int screen_rows = 24;
static int screen_cols = 80;
static int screen_orow = 0;
static int screen_ocol = 0;

static void
goto_origin (int row, int col)
{
  screen_orow = row - 1;
  screen_ocol = col - 1;
}

static void
goto_centerwin (int rows, int cols)
{
  screen_orow = screen_rows/2 - rows/2;
  screen_ocol = screen_cols/2 - cols/2;
}

static void
gotorc (int row, int col)
{
  printf("\033[%d;%dH", row + screen_orow, col + screen_ocol);
}

static void
center_rc (int row, int col, char *s)
{
  gotorc (row, col - strlen(s)/2);
  printf("%s", s);
}

static void
clrscr ()
{
  screen_orow = screen_ocol = 0;
  printf("\033[1;1H\033[2J");
}

#define title(s) clrscr(); center_rc (1, screen_cols/2, s)
#define subtitle(s) center_rc (3, screen_cols/2, s)

/* known special keys

   Ctrl-F = forward (left arrow)	ESC [ C
   Ctrl-B = back (right arrow)		ESC [ D
   Ctrl-N = next line (down arrow)	ESC [ B
   Ctrl-P = prev line (up arrow)	ESC [ A
   Ctrl-A = Home			ESC [ H
   Ctrl-E = End				ESC [ F

   CSI 2 ~ = Insert
   CSI 3 ~ = Delete
   CSI 5 ~ = Page Up
   CSI 6 ~ = Page Down
*/

#define KEY_NONE	0x100
#define KEY_ESC		0x01b
#define KEY_TAB		'\t'
#define KEY_ENTER	'\n'
#define KEY_DELETE	'\b'
#define KEY_UP		0x101
#define KEY_DOWN	0x102
#define KEY_LEFT	0x103
#define KEY_RIGHT	0x104
#define KEY_HOME	0x105
#define KEY_END		0x106
#define KEY_PAGEUP	0x107
#define KEY_PAGEDOWN	0x108
#define KEY_INSERT	0x109
#define KEY_TIMEOUT	0x110
#define KEY_CONNECT	12 /* ctrl-L */

static int key_timeout = 0;

static int
key_low (void)
{
  long now;

  if (key_timeout == 0)
    while (1)
      {
	int k = tty_getc ();
	if (k != 3)
	  return k;
      }
  now = msec;
  while (msec - now < key_timeout)
    {
      if (U1_DTR == 0) /* DTR is like nRESET, normally HIGH when ready */
	{
	  tty_discard_all ();
	  led_color (led_green);
	  while (U1_DTR == 0)
	    ;
	  led_off ();
	  wait_msec (10);
	  tty_init ();
	  wait_msec (10);
	  return KEY_CONNECT;
	}
      if (tty_ready ())
	{
	  int k = tty_getc ();
	  if (k != 3)
	    return k;
	}
    }
  return KEY_TIMEOUT;
}

#define GETKEY(v) v = key_low(); if (v == KEY_TIMEOUT) return v

static int
key (void)
{
  long now;
  int k, got_digits, idx, i;
  int vals[10];

 try_again:
  GETKEY (k);

  if (k != KEY_ESC)
    switch (k)
      {
      case 127:	return KEY_DELETE;
      case 1:	return KEY_HOME;
      case 2:	return KEY_LEFT;
      case 5:	return KEY_END;
      case 6:	return KEY_RIGHT;
      case 16:	return KEY_UP;
      case 14:	return KEY_DOWN;
      case '\r':
      case '\n': return KEY_ENTER;
      default:	return k;
      }

  now = msec;
  while (msec - now < 50)
    if (tty_ready ())
      goto esc_more;
  return k;

 esc_more:
  GETKEY (k);

  if (k == 'O')
    {
      GETKEY (k);
      switch (k)
	{
	default:
	  goto try_again;
	}
    }

  if (k != '[')
    goto try_again;

  idx = -1;
  got_digits = 0;

  while (1)
    {
      GETKEY (k);
      if (isdigit (k))
	{
	  if (!got_digits)
	    {
	      idx ++;
	      vals[idx] = 0;
	      got_digits = 1;
	    }
	  vals[idx] *= 10;
	  vals[idx] += k - '0';
	}
      else if (k == ';')
	{
	  got_digits = 0;
	}
      else
	{
	  switch (k)
	    {
	    case 'A': return KEY_UP;
	    case 'B': return KEY_DOWN;
	    case 'C': return KEY_RIGHT;
	    case 'D': return KEY_LEFT;
	    case 'H': return KEY_HOME;
	    case 'F': return KEY_END;
	    case '~':
	      switch (vals[0])
		{
		case 2: return KEY_INSERT;
		case 3: return KEY_DELETE;
		case 5: return KEY_PAGEUP;
		case 6: return KEY_PAGEDOWN;
		default:
		  goto try_again;
		}

	    case 'R':
	      screen_rows = vals[0];
	      screen_cols = vals[1];
	      return KEY_NONE;

	    default:
	      goto try_again;
	    }
	}
    }
}

#define GL_ANY		0x00
#define GL_DIGITS	0x01
#define GL_SIGN		0x02
#define GL_INT		0x03
#define GL_DECIMAL	0x04
#define GL_FLOAT	0x07

static int
get_line (char *buf, int bufsz, int flags)
{
  int bp = strlen(buf);
  int k;

  tty_puts (buf);

  while (1)
    {
      k = key ();
      /* printf(" \033[34m%03x \033[0m\b\b\b\b\b", k); */

      if (k == KEY_DELETE)
	{
	  if (bp == 0)
	    tty_putc(7);
	  else
	    {
	      tty_puts("\b \b");
	      bp--;
	    }
	  continue;
	}

      if (k == KEY_ENTER)
	{
	  buf[bp] = 0;
	  return KEY_ENTER;
	}

      if (k <= 0 || k >= 0x100 || !isprint (k))
	{
	  return k;
	}

      if (flags)
	{
	  if (!(   (flags & GL_DIGITS  && isdigit (k))
		|| (flags & GL_SIGN    && ( k=='+' || k=='-' ))
		|| (flags & GL_DECIMAL && ( k=='.' ))
		))
	    {
	      tty_putc (7);
	      continue;
	    }
	}

      if (bp == bufsz - 1)
	{
	  tty_putc (7);
	  continue;
	}

      tty_putc (k);
      buf[bp++] = k;
    }
}

static int
fetch_float_1 (int r, int c, const char *msg, char *buf, int buflen, float *ret)
{
  float rv = *ret;

  if (rv == 0)
    buf[0] = 0;
  else
    sprintf (buf, "%g", rv);

  gotorc (r, c);
  tty_puts (msg);
  c += strlen (msg);

  while (1)
    {
      char *eptr;
      gotorc (r, c);
      tty_puts("\033[K");
      if (get_line (buf, buflen, GL_DIGITS|GL_DECIMAL) == KEY_ESC)
	return 1;
      rv = strtof (buf, &eptr);
      if (rv >= 0 && eptr != buf)
	{
	  *ret = rv;
	  return 0;
	}
      tty_puts(7);
    }
}

#define fetch_float(r,msg,buf,ret) if (fetch_float_1 (r, 1, msg, buf, sizeof(buf), &ret)) return;

/*====================================================================== */

#define KEY_CHECK if (tty_ready()) { motor_stop (); return; }

static void
commutate_to (int u, int v, int w)
{
  int power = 0;
  long start = msec;
  int mtick = motor_pwm_ticks;
  int its_moving = 0;
  int max_power = 32767;
  int max_rpm = 0;
  int rpm;
  int overamp_counts = 0;
  int underamp_counts = 0;

  printf("\033[34mcommutate to %d %d %d\033[0m\n", u, v, w);
  while (1)
    {
      /* we allow up to 1 second for the motor to move  */
      if (msec - start > 1000)
	{
	  printf("\033[33mtimeout, must already be there?\033[0m\n");
	  break;
	}

      /* The power level slowly increases over time, from 0 to 100% */
      power = (msec - start) * 32768 / 1000;
      if (power > max_power)
	power = max_power;

      /* We use up to half an amp, it's small but safe.  */
      if (motor_current > 0.5)
	{
	  cpwm_levels (0, 0, 0);
	  if (overamp_counts == 0)
	    printf("\033[31mamp limit reached\033[0m\n");
	  overamp_counts ++;
	  max_power = power - 1;
	}
      else
	{
	  cpwm_levels (u*power, v*power, w*power);
	  underamp_counts ++;
	}

      while (motor_pwm_ticks == mtick)
	;
      mtick = motor_pwm_ticks;

      rpm = encoder_rpm ();
      if (rpm > 100)
	{
	  if (!its_moving)
	    {
	      printf("\033[32mit's moving at power %d\033[0m\n", power);
	    }
	  its_moving = 1;
	  max_power = power;
	}
      if (its_moving && rpm < 10)
	break;
      if (max_rpm < rpm)
	max_rpm = rpm;

#if 0
      if (tty_idle())
	{
	  printf("  %d %d %d %d  \r", power, rpm, its_moving, max_power);
	}
#endif
    }

  start = msec;
  while (1)
    {
      /* steady decrease to stop at best spot */
      power = max_power - (msec - start) * 32768 / 1000;
      if (power < 0)
	break;
      if (motor_current > 0.5)
	cpwm_levels (0, 0, 0);
      else
	cpwm_levels (u*power, v*power, w*power);
      while (motor_pwm_ticks == mtick)
	;
      mtick = motor_pwm_ticks;
    }

  wait_msec (100);
  cpwm_levels (0, 0, 0);
#if 0
  printf("max power: %d, max_rpm: %d (end %d)   over/under %d/%d\n",
	 max_power, max_rpm, rpm, overamp_counts, underamp_counts);
#endif
}

static int
double_commutate (int u, int v, int w, int nu, int nv, int nw)
{
  int a, b;
  commutate_to (u, v, w);
  a = encoder_val ();
  commutate_to (nu, nv, nw);
  commutate_to (u, v, w);
  b = encoder_val ();
  a = ((a+b)/2) % 2000;
  a *= 12;
  a = ((int)((a+500)/1000))*1000;
  a /= 12;
  return a;
}

static char *bits(int h)
{
  static char rv[2][4];
  static int ri = 0;
  ri = 1-ri;
  rv[ri][0] = (h & U_HALL) ? 'U' : '-';
  rv[ri][1] = (h & V_HALL) ? 'V' : '-';
  rv[ri][2] = (h & W_HALL) ? 'W' : '-';
  rv[ri][3] = 0;
  return rv[ri];
}

static void
do_config_wiring ()
{
  int k;
  Config_Wiring cfg;
  int mins[16], maxs[16];
  Position startp, ep, en;
  int old_h;
  long start_msec;

  memset (&cfg, 0, sizeof(cfg));

  title ("BLDC Configuration Menu - Axis");
  gotorc (3, 1);

  motor_calibrate ();
  bldc_run ();

  while (!encoder_synced ())
    {
      printf("\033[35msyncing encoder...\033[0m\n");
      commutate_to (+1, -1, -1);
      commutate_to (-1, -1, +1);
      commutate_to (-1, +1, -1);
    }

  cfg.encoder_at_u = double_commutate (+1, -1, -1, -1, -1, +1);
  printf("\033[36mEncoder at U = %d\033[0m\n", cfg.encoder_at_u);

  cfg.encoder_at_v = double_commutate (-1, +1, -1, -1, -1, +1);
  printf("\033[36mEncoder at V = %d\033[0m\n", cfg.encoder_at_v);

  cfg.encoder_at_w = double_commutate (-1, -1, +1, +1, -1, -1);
  printf("\033[36mEncoder at W = %d\033[0m\n", cfg.encoder_at_w);

  /* Determine winding direction.  */
  k = cfg.encoder_at_v - cfg.encoder_at_u;
  while (k > 2000)
    k -= 2000;
  while (k < -2000)
    k += 2000;
  if (k > 0)
    cfg.sincos_negate = 1;
  else
    cfg.sincos_negate = -1;

  /* We save this so that encoder_ang() basically works, which RPM
     mode needs.  We'll save it again later */
  config_wiring_save (&cfg);

  /* Now, run the motor and scan the hall sensors.  */
  printf("start lval %lld %02x\n", encoder_lval (), HALLS);
  motor_rpm (100);

  memset (mins, ~0, sizeof(mins));
  memset (maxs,  0, sizeof(maxs));
  startp = encoder_lval ();
  ep = startp + 4000;
  en = startp - 4000;

  start_msec = msec;
  old_h = HALLS & (U_HALL | V_HALL | W_HALL);
  while (1)
    {
      Position here = encoder_lval ();
      int h;
      Position v;

      if (msec - start_msec > 3000)
	{
	  printf("timeout!  try again, pelase\n");
	  motor_stop ();
	  bldc_stop ();
	  key();
	  return;
	}

      if (here < en || here > ep)
	break;

      h = HALLS & (U_HALL | V_HALL | W_HALL);
      if (h == old_h)
	continue;

      v = encoder_val () % 2000;
      /* printf("%s -> %s at %d\n", bits(old_h), bits(h), v); */
      maxs[old_h] = v;
      mins[h] = v;
      old_h = h;
    }
  motor_stop ();
  printf("end   lval %lld %02x \n\n", encoder_lval (), HALLS);

  for (k=2; k<14; k++)
    if ((k & ~(U_HALL | V_HALL | W_HALL)) == 0)
      {
	int a = maxs[k];
	if (maxs[k] < mins[k])
	  a += 2000;
	a = (mins[k] + a)/2;
	a *= 12;
	a = ((int)((a+250)/500))*500;
	a /= 12;
	printf("%3s = %4d - %4d  (%4d)\n", bits(k), mins[k], maxs[k], a);
	cfg.halls_to_encoder[k] = a;
      }

  bldc_stop ();

  config_wiring_save (&cfg);

  printf ("\ndone!\n");
  key();

}

static void
do_config_axis ()
{
  int k;
  Config_Axis cfg;

  if (config_axis_done ())
    memcpy (&cfg, config_axis_get (), sizeof(Config_Axis));
  else
    memset (&cfg, 0, sizeof(cfg));

  title ("BLDC Configuration Menu - Axis");

  goto_centerwin (1, sizeof("Choose UART2 Axis (A,B,C,Del): [x]   "));

  gotorc (1, 1);
  tty_puts("Choose Axis (X,Y,Z): ");
  if (cfg.my_axis)
    printf("[%c] ", cfg.my_axis);
  while (1)
    {
      k = key ();
      if ((k == KEY_TAB || k == KEY_ENTER) && cfg.my_axis)
	break;
      else if (isalpha (k))
	{
	  cfg.my_axis = toupper (k);
	  break;
	}
    }
  tty_putc (cfg.my_axis);

  gotorc (2, 1);
  tty_puts("Choose UART2 Axis (A,B,C,Del): ");
  if (cfg.uart2_axis)
    printf("[%c] ", cfg.uart2_axis);
  while (1)
    {
      k = key ();
      if (k == KEY_TAB || k == KEY_ENTER)
	break;
      else if (k == KEY_DELETE)
	{
	  cfg.uart2_axis = 0;
	  break;
	}
      else if (isalpha (k))
	{
	  cfg.uart2_axis = toupper (k);
	  break;
	}
    }
  tty_putc (cfg.uart2_axis);

  config_axis_save (&cfg);
}

static const char * const set_str[] = {
  "LOW ", "YES", "YES", "YES", "YES", "YES", "YES" };
static const char * const clr_str[] = {
  "HIGH", "---", "---", "---", "---", "---", "---" };
static const char * const id_str[] = {
  "active when",
  "positive limit",
  "negative limit",
  "positive home",
  "negative home",
  "LIMIT1 output",
  "LIMIT2 output"
};

static void
do_config_limits ()
{
  int k, i, j;
  int r = 0, c = 1;
  Config_Limits cfg;
  Config_Limits cfg_seen;
  unsigned char *cp = (unsigned char *) &cfg;
  unsigned char *cps = (unsigned char *) &cfg_seen;
  int old_lim = -1, lim;
  int rpm = 60;
  int motor_running = 0;
  char *err = NULL;

  motor_stop ();

  /* We take advantage of the fact that the config is a simple list of
     bytes, one for each purpose, with bits that match the LIMIT* bit
     positions (i.e. LIMIT3 is 1<<3 etc) */

  if (config_limits_done ())
    memcpy (&cfg, config_limits_get (), sizeof(Config_Limits));
  else
    {
      memset (&cfg, 0, sizeof(cfg));
      /* assume no switches active at this time */
      cfg.polarity = LIMITS;
      cfg.limit_neg = 0x12;
      cfg.limit_pos = 0x12;
      cfg.limit1 = 0x12;
      cfg.limit2 = 0x12;
    }

  for (i=0; i<sizeof(cfg); i++)
    cps[i] = ~cp[i];

  title ("BLDC Configuration Menu - Limit Switches");
  subtitle ("arrow keys to select, SPACE to toggle, ENTER when done");

  goto_centerwin (13, sizeof("[P] or [N] to move motor towards positive/negative limits"));

  gotorc (1,5);
  tty_puts ("LIM1    LIM2    LIM3    LIM4");

  gotorc (12,1);
  tty_puts ("[P] or [N] to move motor towards positive/negative limits");
  gotorc (13,1);
  tty_puts ("1..9 for speed, [S] stops   [R] reads current active state");

  for (j=0; j<7; j++)
    {
      gotorc (4+j, 5*8-3);
      tty_puts (id_str[j]);
    }

  key_timeout = 100;

  while (1)
    {
      gotorc (13, 1);
      if (err)
	{
	  printf("\033[31m%s\033[0m", err);
	  err = NULL;
	}
      else
	printf("\033[K");

      for (i=1; i<=4; i++)
	for (j=0; j<7; j++)
	  {
	    if ((cp[j] & (1<<i)) != (cps[j] & (1<<i)))
	      {
		gotorc (4+j, i*8-3);
		tty_puts((cp[j] & (1<<i)) ? set_str[j] : clr_str[j]);
	      }
	  }
      memcpy (cps, cp, sizeof(cfg));

    redraw_limits:
#define SLIM(i)						\
      gotorc (2, i*8-3);				\
      if (LIMIT##i) tty_puts("\033[32mHIGH\033[0m");	\
      else          tty_puts("\033[31mLOW \033[0m");
    SLIM (1);
    SLIM (2);
    SLIM (3);
    SLIM (4);
    old_lim = LIMITS;


    fflush (stdout);
    gotorc (4+r, c*8-5);
    tty_putc('>');
    keep_waiting:
    k = key ();
    gotorc (4+r, c*8-5);
    if (k == KEY_TIMEOUT)
      {
	lim = LIMITS;
	if (lim == old_lim)
	  goto keep_waiting;
      }	  
	  
    tty_putc(' ');
    if (k == KEY_TIMEOUT)
      goto redraw_limits;

#define MNOTE(s) gotorc (2, 5*8-3); tty_puts(s)
#define MSTOP { motor_running = 0; MNOTE("\033[K"); motor_stop (); }

    switch (k)
      {
      case KEY_ENTER:
	if (! cfg.limit_pos || ! cfg.limit_neg)
	  {
	    err = "missing pos or neg limit";
	    break;
	  }
	config_limits_save (&cfg);
      case KEY_ESC:
	key_timeout = 0;
	MSTOP;
	return;

      case KEY_UP:
	if (r > 0)
	  r --;
	break;
      case KEY_DOWN:
	if (r < 6)
	  r ++;
	break;
      case KEY_LEFT:
	if (c > 1)
	  c --;
	break;
      case KEY_RIGHT:
	if (c < 4)
	  c ++;
	break;

      case ' ':
	motor_stop ();
	cp[r] ^= (1<<c);
	break;

      case 'r':
      case 'R':
	cfg.polarity = LIMITS;
	MSTOP;
	break;

      case 'p':
      case 'P':
	if (motor_running == 1)
	  MSTOP
	  else
	    {
	      motor_running = 1;
	      MNOTE("To Positive");
	      motor_rpm (rpm);
	    }
	break;
      case 'n':
      case 'N':
	if (motor_running == -1)
	  MSTOP
	  else
	    {
	      motor_running = -1;
	      MNOTE("To Negative");
	      motor_rpm (-rpm);
	    }
	break;
      case '0':
	rpm = 60;
      case 's':
      case 'S':
	MSTOP
	  break;

      }
    if (k >= '1' && k <= '9')
      {
	rpm = (k - '0') * 60;
	if (motor_running)
	  motor_rpm (motor_running * rpm);
      }
    }
}

static void
do_config_steps ()
{
  int k;
  Config_Steps cfg;
  int per_rot;
  char buf[20];
  char *err = NULL;

  if (config_steps_done ())
    memcpy (&cfg, config_steps_get (), sizeof(Config_Steps));
  else
    {
      memset (&cfg, 0, sizeof(cfg));
      cfg.position_change_per_step = MAXC / 2000;
    }

  if (MAXC % cfg.position_change_per_step == 0)
    per_rot = 1;
  else
    per_rot = 0;

  while (1)
    {
      title ("BLDC Configuration Menu - Step Rate");
#define ESTR2(x) # x
#define ESTR(x) ESTR2(x)
      subtitle (ESTR(MAXC) " encoder steps per rotation");

      goto_centerwin (3, sizeof("Steps per rotation: 123"));
      gotorc (3, 1);
      tty_puts("TAB toggles mode, ENTER when done");
      gotorc (4, 1);
      if (err)
	{
	  tty_puts("\033[31m");
	  tty_puts(err);
	  tty_puts("\033[0m");
	  err = NULL;
	}
      else
	tty_puts("\033[K");

      if (per_rot)
	{
	  gotorc (1, 1);
	  tty_puts("Steps per rotation: ");
	  sprintf (buf, "%d", MAXC / cfg.position_change_per_step);
	  k = get_line (buf, sizeof(buf), GL_INT);
	  switch (k)
	    {
	    case KEY_TAB:
	      per_rot = ! per_rot;
	      break;
	    case KEY_ESC:
	      return;
	    case KEY_ENTER:
	      k = atoi (buf);
	      if (abs(k) < 4)
		{
		  err = "value too small (<4)";
		  break;
		}
	      if (abs(k) > MAXC)
		{
		  err = "value too big (>" ESTR(MAXC) ")";
		  break;
		}
	      if (MAXC % k != 0)
		{
		  err = "value not divisible into 4000";
		  break;
		}
	      cfg.position_change_per_step = MAXC / k;
	      goto step_config_done;
	    }
	}
      else
	{
	  gotorc (1, 1);
	  tty_puts("Encoder counts per step: ");
	  sprintf (buf, "%d", cfg.position_change_per_step);
	  k = get_line (buf, sizeof(buf), GL_INT);
	  switch (k)
	    {
	    case KEY_TAB:
	      per_rot = ! per_rot;
	      break;
	    case KEY_ESC:
	      return;
	    case KEY_ENTER:
	      k = atoi (buf);
	      if (k == 0)
		{
		  err = "value must not be zero";
		  break;
		}
	      if (abs(k) > MAXC)
		{
		  err = "value too big (>" ESTR(MAXC) ")" ;
		  break;
		}
	      cfg.position_change_per_step = k;
	      goto step_config_done;
	    }
	}
    }
 step_config_done:

  config_steps_save (&cfg);
}

do_config_homing_rpm()
{
  int k;
  Config_Home_RPM cfg;
  char *err = NULL;
  char buf[10];

  if (config_home_rpm_done ())
    memcpy (&cfg, config_home_rpm_get (), sizeof(Config_Home_RPM));
  else
    {
      memset (&cfg, 0, sizeof(cfg));
      cfg.rpm = 600;
    }

  while (1)
    {
      title ("BLDC Configuration Menu - Step Rate");
      goto_centerwin (3, sizeof("Homing RPM: 123"));

      gotorc (3, 1);
      if (err)
	{
	  tty_puts("\033[31m");
	  tty_puts(err);
	  tty_puts("\033[0m");
	  err = NULL;
	}
      else
	tty_puts("\033[K");

      gotorc (1, 1);
      tty_puts("Homing RPM: ");

      sprintf (buf, "%d", cfg.rpm);
      k = get_line (buf, sizeof(buf), GL_DIGITS);
      switch (k)
	{
	case KEY_ESC:
	  return;
	case KEY_ENTER:
	  k = atoi (buf);
	  if (k < 1)
	    {
	      err = "value must be positive";
	      break;
	    }
	  if (k > 60000)
	    {
	      err = "you're kidding, right?";
	      break;
	    }
	  cfg.rpm = k;
	  config_home_rpm_save (&cfg);
	  return;
	}

    }
}

/* This assumes the config data is a list of floats, and is indexed
   the same as the float data layout.  */
static const char * const motor_fields[] = {
  "Winding Resistrance (ohms)   : ",
  "Winding Inductance (henries) : ",
  "Volts/kRPM                   : ",
  "Max Continuous Amps          : ",
  "Max Peak Amps                : ",
  "Max Volts                    : ",
  "Position PID P               : ",
  "Position PID I               : ",
  "Position PID Imax            : ",
  "Position PID FF              : ",
  "RPM PID P                    : ",
  "RPM PID I                    : ",
  "RPM PID Imax                 : ",
  "RPM PID FF                   : ",
  "Current Q PID P              : ",
  "Current Q PID I              : ",
  "Current Q PID Imax           : ",
  "Current Q PID FF             : ",
  "Current D PID P              : ",
  "Current D PID I              : ",
  "Current D PID Imax           : ",
  "Current D PID FF             : ",
};
#define N_MOTOR_FIELDS (sizeof(motor_fields)/(sizeof(motor_fields[0])))

do_config_motor_calib ()
{
  int i, k;
  Config_Motor_Calib cfg;
  float *cfg_f;
  char buf[20];

  if (config_motor_calib_done ())
    memcpy (&cfg, config_motor_calib_get (), sizeof(Config_Motor_Calib));
  else
    {
    set_to_defaults:
      memset (&cfg, 0, sizeof(cfg));
      cfg.winding_ohms = 5.6;
      cfg.winding_henries = 0.0082;
      cfg.volts_per_krpm = 35.1;
      cfg.max_amps = 2.55;
      cfg.peak_amps = 7.4;
      cfg.max_volts = 60;

      cfg.pos_pid.p    = 1;
      cfg.pos_pid.i    = 100;
      cfg.pos_pid.imax = 50;
      cfg.pos_pid.ff   = 0.5;

      cfg.rpm_pid.p    = 0.02;
      cfg.rpm_pid.i    = 1;
      cfg.rpm_pid.imax = 3;
      cfg.rpm_pid.ff   = 0.003;

      cfg.current_q_pid.p    = 0.25;
      cfg.current_q_pid.i    = 0;
      cfg.current_q_pid.imax = 0;
      cfg.current_q_pid.ff   = 0;

      cfg.current_d_pid.p    = 0.25;
      cfg.current_d_pid.i    = 0;
      cfg.current_d_pid.imax = 0;
      cfg.current_d_pid.ff   = 0;
    }

  cfg_f = (float *)&cfg;
  title ("BLDC Configuration Menu - Motor Calibration");

  goto_centerwin (7, sizeof("Winding Resistance (ohms)    : 0.0000"));
  for (i=0; i<N_MOTOR_FIELDS; i++)
    {
      gotorc (i+1, 1);
      tty_putc ('A' + i);
      tty_puts (". ");
      tty_puts(motor_fields[i]);
      sprintf (buf, "%g", cfg_f[i]);
      tty_puts(buf);
      tty_puts("\033[K");
    }
  gotorc (N_MOTOR_FIELDS+3, 1);
  tty_puts("'=' to load defaults");
  gotorc (N_MOTOR_FIELDS+2, 1);
  tty_puts("Select, or ESC to cancel, ENTER to save : ");
  while (1)
    {
      gotorc (N_MOTOR_FIELDS+2, 43);
      k = tolower (key ());
      if ('a' <= k && k < 'a'+N_MOTOR_FIELDS)
	{
	  i = k - 'a';
	  fetch_float_1 (i+1, 4+ strlen(motor_fields[0]), "", buf, sizeof(buf), cfg_f+i);
	}
      if (k == KEY_ESC)
	return;
      if (k == KEY_ENTER)
	break;
      if (k == '=')
	goto set_to_defaults;
    }

  config_motor_calib_save (&cfg);
}

static void
do_calibrate_position_loop ()
{
  Position start_pos;
  Position pos_pos, neg_pos;
  Position delta;
  int sign;
  int k;
  long start_msec;

  title ("BLDC Configuration Menu - RPM Loop Calibration");

  if (!config_limits_done ())
    {
      gotorc (3,3);
      tty_puts ("Need to config limits first\n");
      key();
      return;
    }

  if (!config_motor_calib_done ())
    {
      gotorc (3,3);
      tty_puts ("Need to config motor first\n");
      key();
      return;
    }

  start_pos = encoder_lval ();
  calib_target_position = start_pos;

  motor_travel (start_pos);

  sign = 1;
  delta = 1;
  start_msec = msec;
  while (1)
    {
      if (msec - start_msec > 1000)
	{
	  start_msec = msec;
	  sign = -sign;
	  if (sign > 0)
	    tty_putc('+');
	  else
	    tty_putc('-');
	  motor_travel (start_pos + sign * delta);
	}

      if (tty_ready ())
	{
	  k = key ();
	  switch (k)
	    {
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
	      delta = config_steps.position_change_per_step << (k - '0');
	      printf(" new delta %lld\n", delta);
	      break;
	    case KEY_ESC:
	      goto calib_done;
	    }
	}

    }
 calib_done:
  motor_stop ();
  while (tty_ready()) key();
  printf("done! "); key();
}

static void
do_calibrate_rpm_loop ()
{
  Position start_pos;
  Position pos_pos, neg_pos;
  float rpm;
  int sign;
  int k;
  long start_msec;

  title ("BLDC Configuration Menu - RPM Loop Calibration");

  if (!config_limits_done ())
    {
      gotorc (3,3);
      tty_puts ("Need to config limits first\n");
      key();
      return;
    }

  if (!config_motor_calib_done ())
    {
      gotorc (3,3);
      tty_puts ("Need to config motor first\n");
      key();
      return;
    }

  start_pos = encoder_lval ();
  pos_pos = start_pos + 2000;
  neg_pos = start_pos - 2000;

  motor_calibrate_rpm_loop ();

  sign = 1;
  rpm = 0.0;
  start_msec = msec;
  while (1)
    {
      Position here = encoder_lval ();
      if (here < neg_pos && sign != 1)
	{
	  tty_putc('+');
	  sign = 1;
	  start_msec = msec;
	}
      if (here > pos_pos && sign != -1)
	{
	  tty_putc('-');
	  sign = -1;
	  start_msec = msec;
	}
      if (msec - start_msec > 1000)
	{
	  tty_putc('*');
	  sign = -sign;
	  start_msec = msec;
	}

      if (tty_ready ())
	{
	  k = key ();
	  switch (k)
	    {
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
	      rpm = 100 * (k - '0');
	      printf(" new rpm %f\n", rpm);
	      break;
	    case KEY_ESC:
	      goto calib_done;
	    }
	}

      calib_desired_rpm = sign * rpm;
    }
 calib_done:
  motor_stop ();
  while (tty_ready()) key();
  printf("done! "); key();
}

static void
do_calibrate_current_loop ()
{
  Position start_pos;
  Position pos_pos, neg_pos;
  float current;
  int sign;
  int k;
  long start_msec;

  title ("BLDC Configuration Menu - Current Loop Calibration");

  if (!config_limits_done ())
    {
      gotorc (3,3);
      tty_puts ("Need to config limits first\n");
      key();
      return;
    }

  if (!config_motor_calib_done ())
    {
      gotorc (3,3);
      tty_puts ("Need to config motor first\n");
      key();
      return;
    }

  start_pos = encoder_lval ();
  pos_pos = start_pos + 2000;
  neg_pos = start_pos - 2000;

  motor_calibrate_current_loop ();

  sign = 1;
  current = 0.0;
  start_msec = msec;
  while (1)
    {
      Position here = encoder_lval ();
      if (here < neg_pos && sign != 1)
	{
	  tty_putc('+');
	  sign = 1;
	  start_msec = msec;
	}
      if (here > pos_pos && sign != -1)
	{
	  tty_putc('-');
	  sign = -1;
	  start_msec = msec;
	}
      if (msec - start_msec > 1000)
	{
	  tty_putc('*');
	  sign = -sign;
	  start_msec = msec;
	}

      if (tty_ready ())
	{
	  k = key ();
	  switch (k)
	    {
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
	      current = 0.1 * (k - '0');
	      printf(" current %f\n", current);
	      break;
	    case KEY_ESC:
	      goto calib_done;
	    }
	}

      calib_desired_current = sign * current;
    }
 calib_done:
  motor_stop ();
  while (tty_ready()) key();
  printf("done! "); key();
}

static void
do_calibrate_sine_test ()
{
  title ("BLDC Configuration Menu");
  int range, speed;
  int new_range, new_speed;
  Position center;
  int dpos_idx;
  int mtick;
  int old_sign = 1;
  Position cur_pos, old_pos;

  goto_centerwin (4, sizeof("0-9 : change speed"));
  gotorc (1, 1); tty_puts("0-9 : change range");
  gotorc (2, 1); tty_puts("A-Z : change speed");
  gotorc (3, 1); tty_puts("ESC to stop");
  gotorc (4, 1); tty_puts("> ");

  /* One dpos step per pwm cycle is the fastest we go; the slowest is
     1/10th that.  I.e.  we increment dpos_idx every SPEED mticks.  */

  new_range = range = 0; /* smallest range */
  new_speed = speed = 26; /* slowest speed */
  dpos_idx = 0;
  old_pos = center = encoder_lval ();
  mtick = motor_pwm_ticks;
  calib_target_position = center;

  motor_travel (center);
  while (1)
    {
      if (tty_ready ())
	{
	  int k = key ();
	  if (k == KEY_ESC)
	    break;

	  if ((k >= 'a' && k <= 'z')
	      || (k >= 'Z' && k <= 'Z'))
	    {
	      new_speed = tolower (k) - 'a' + 1;
	    }
	  if (k >= '0' && k <= '9')
	    {
	      if (k == '0')
		k = '9' + 1;
	      new_range = (k - '0') * (MAXC/2/10);
	    }
	  gotorc (4, 2);
	  printf("R:%4d S:%2d ", new_range, new_speed);
	  fflush(stdout);
	}

      if ((motor_pwm_ticks - mtick) > speed)
	{
	  mtick += speed;
	  dpos_idx ++;
	  old_pos = cur_pos;
	  cur_pos = dcos (dpos_idx) * range + center;
	  motor_travel (cur_pos);

	  if (((cur_pos < center) != (old_pos < center)) || old_pos == center)
	    {
	      speed = new_speed;
	      range = new_range;
	    }
	}
    }
  motor_stop ();
  bldc_stop ();
}

static int always_done() { return 1; }

static void
do_config_menu ()
{
  while (1)
    {
    start_again:
      title ("BLDC Configuration Menu");

      goto_centerwin (15, sizeof("1. Check Wiring & Phase [needed]"));

#define CM(r,s,f)				\
      gotorc(r,1); tty_puts (s);		\
      if (!f ()) tty_puts(" [needed]");

      CM(1, "1. Check Wiring & Phase", config_wiring_done);
      CM(2, "2. Set Axis", config_axis_done);
      CM(3, "3. Limit Switches", config_limits_done);
      CM(4, "4. Set Gearing", config_gearing_done);
      CM(5, "5. Steps per Rotation", config_steps_done);
      CM(6, "6. Homing RPM", config_home_rpm_done);
      CM(7, "7. UART 2", config_uart2_done);
      CM(8, "8. Motor Calibration", config_motor_calib_done);
      CM(10, "A. Calibrate Position Loop", always_done);
      CM(11, "B. Calibrate Current Loop", always_done);
      CM(12, "C. Calibrate Current Loop", always_done);
      CM(13, "S. Sine Wave Test", always_done);

      gotorc (15, 1);
      tty_puts("Choose: ");
      goto_origin (0, 0);

      motor_stop ();
      switch (key())
	{
	case KEY_CONNECT:
	  goto start_again;
	case '1':
	  do_config_wiring ();
	  break;
	case '2':
	  do_config_axis ();
	  break;
	case '3':
	  do_config_limits ();
	  break;
	case '5':
	  do_config_steps ();
	  break;
	case '6':
	  do_config_homing_rpm ();
	  break;
	case '8':
	  do_config_motor_calib ();
	  break;
	case 'a':
	case 'A':
	  do_calibrate_position_loop ();
	  break;
	case 'b':
	case 'B':
	  do_calibrate_rpm_loop ();
	  break;
	case 'c':
	case 'C':
	  do_calibrate_current_loop ();
	  break;
	case 's':
	case 'S':
	  do_calibrate_sine_test ();
	  break;
	case KEY_ESC:
	  return;
	}
    }
}

extern volatile Position target_position;

static void
do_main_menu ()
{
  long timeout;
  Config_Steps *cfg_step = config_steps_get ();
  Config_Limits *cfg_limits = config_limits_get ();
  Position here;
  Position off_by;

  title ("BLDC Main Menu");

  goto_centerwin (4, sizeof("<[ ]>  step <>=1 []=10"));
  gotorc (1, 1); tty_puts ("<[ ]>  Step <>=1 []=10");
  gotorc (2, 1); tty_puts ("C      Configuration menu");
  gotorc (3, 1); tty_puts ("ESC    Stop");
  gotorc (4, 1); tty_puts ("R      Start");
  goto_origin (1,1);

  while (1)
    {
      timeout = msec + 200;
      while (msec - timeout < 0)
	if (tty_ready ())
	  {
	    switch (key ())
	      {
	      case KEY_ESC:
		motor_stop ();
		break;

	      case KEY_CONNECT:
		return;

	      case 'r':
	      case 'R':
		motor_track ();
		break;

	      case 'c':
	      case 'C':
		motor_stop ();
		do_config_menu ();
		return;

	      case '<':
	      case ',':
		motor_step (-cfg_step->position_change_per_step);
		break;
	      case '>':
	      case '.':
		motor_step (cfg_step->position_change_per_step);
		break;
	      case '[':
		motor_step (-cfg_step->position_change_per_step * 10);
		break;
	      case ']':
		motor_step (cfg_step->position_change_per_step * 10);
		break;
	      case '{':
		motor_step (-cfg_step->position_change_per_step * 100);
		break;
	      case '}':
		motor_step (cfg_step->position_change_per_step * 100);
		break;
	      }
	  }

      gotorc (2, 1);
      here = encoder_lval ();
      off_by = here - target_position;
      printf("Position: %10lld ", here);
      if (off_by > 0) {
	printf("\033[31m");
      } else if (off_by < 0) {
	printf("\033[34m");
      } else {
	printf("\033[32m");
      }

      extern volatile float desired_rpm;
      extern volatile float desired_current;

      printf("vs  %10lld (%+d)   ", target_position, here - target_position);
      printf("\033[0m");
      gotorc (3,1);
      printf("Power  %3d V,  %8.5f A ", (int)motor_volts (), motor_current);
      printf("%04x %04x %04x ", UCURR, VCURR, WCURR);
      gotorc (4,1);
      printf("Temp: %5.1f F ", temperature ());
      gotorc (5, 1);
      printf("Limits: ");
      printf(NLIMIT_HIT() ? "NL" : "--");
      printf(NHOME_HIT() ? "NH" : "--");
      printf(PHOME_HIT() ? "PH" : "--");
      printf(PLIMIT_HIT() ? "PS" : "--");
      gotorc (6, 1);
      printf("Torque: %6d   RPM%6d  \n", cpwm_torque_applied, motor_pwm_rpm);
      motor_status();
      extern int usign, vsign, wsign;
      printf("sign %2d %2d %2d\033[K\n", usign, vsign, wsign);

#if 0
      if (state_n == NSTATE)
	{
	  int i;
	  printf("\nSTART\n");
	  for (i=0; i<state_n; i++)
	    {
	      State *s = state + i;
	      printf("%d %d %d %d %f %f %d %d\n", i, s->off_by, s->rpm, s->d_rpm,
		     s->milliamps / 1000.0, s->d_milliamps / 1000.0,
		     s->torque, s->state);
	    }
	  state_n ++;
	  printf("\nEND\n");
	}
#endif
    }
}

main()
{
  int k;

  bldc_init ();

  wait_msec (20);

    S12AD1.ADANS.WORD = 0x2000;
    S12AD1.ADCSR.BYTE = 0x4c;
    S12AD1.ADCSR.BYTE = 0xcc;
    S12AD1.ADCER.BIT.ADRFMT = 1;
  
#if 0
  while (1)
    {
      printf(" %04x %04x %04x %d %d %d %d \r",
	     UCURR, VCURR, WCURR,
	     LIMIT4, LIMIT3, LIMIT2, LIMIT1
	     );
    }
#endif

  tty_puts("\033[999;999H"); /* move to lower right */
  tty_puts("\033[6n"); /* report cursor position */
  clrscr ();
  fprintf (stderr, "stderr goes here\n");

  led_color (led_red);
  wait_msec (100);
  led_color (led_green);
  wait_msec (100);
  led_color (led_blue);
  wait_msec (100);
  led_color (led_white);
  wait_msec (100);
  led_off ();

  key_timeout = 200;
  while (key() != KEY_TIMEOUT)
    ;
  key_timeout = 0;

  while (!config_valid())
    {
      led_flash (led_blue);
      do_config_menu ();
      led_off ();
    }

  printf("main menu?\n");

  motor_track ();

  while (1)
    do_main_menu ();
}
